# ported from olap4j
# https://github.com/olap4j/olap4j/blob/b8eccd85753ffddb66c9d8b7c2cd7de2bd510ce0/src/org/olap4j/impl/IdentifierParser.java

STATE = type('Enum',
             (),
             dict(START=0, BEFORE_SEG=1, IN_BRACKET_SEG=2, AFTER_SEG=3, IN_SEG=4))

SYNTAX = type('Enum',
             (),
             dict(NAME=0, FIRST_KEY=1, NEXT_KEY=2))

QUOTING = type('Enum',
               (),
               dict(UNQUOTED=0, QUOTED=1, KEY=2))

class Segment(object):
    def __init__(self, name, quoting):
        self._name = name
        self._quoting = quoting

    @property
    def name(self):
        if self._quoting == QUOTING.QUOTED:
            # unquote
            return self._name[1:-1]
        else:
            return self._name

    def __repr__(self):
        return "<Segment \'%s\' quoting=%d>" % (self._name, self._quoting)

class Identifier(object):
    def __init__(self):
        self._segments = []
        self._subsegments = []

    def _flush_subsegments(self):
        if len(self._subsegments) > 0:
            self._segments.append(
                Segment(
                    '&' + '&'.join([ss.name for ss in self._subsegments]),
                    QUOTING.KEY
                )
            )
            self._subsegments = []

    def add_segment(self, string, quoting, syntax):

        segment = Segment(string, quoting)

        if syntax != SYNTAX.NEXT_KEY:
            self._flush_subsegments()

        if syntax == SYNTAX.NAME:
            self._segments.append(segment)
        else:
            self._subsegments.append(segment)

    def __str__(self):
        self._flush_subsegments()
        return '.'.join([s.name for s in self._segments])

    def __getitem__(self, i):
        return self._segments[i]

    @property
    def segments(self):
        self._flush_subsegments()
        return self._segments

    @classmethod
    def parse(cls, string):
        k = len(string)

        pstate = { 'i': 0, 'state': STATE.START, 'start': 0, 'syntax': SYNTAX.NAME }

        ident = cls()

        #ipdb.set_trace()
        def inner():

            c = string[pstate['i']]

            if pstate['state'] in (STATE.START, STATE.BEFORE_SEG):
                if c == '[':
                    pstate.update({
                        'i': pstate['i'] + 1,
                        'start': pstate['i'],
                        'state': STATE.IN_BRACKET_SEG
                    })
                elif c == ' ':
                    pstate['i'] += 1
                elif c in (',', '}', '\0'):
                    return
                elif c == '.':
                    raise ValueError("Unexpected '.'")
                elif c == '&':
                    pstate['i'] += 1
                    if pstate['syntax'] != SYNTAX.NAME:
                        raise ValueError("Unexpected '&'")
                    pstate['syntax'] = SYNTAX.FIRST_KEY
                else:
                    pstate.update({
                        'state': STATE.IN_SEG,
                        'start': pstate['i']
                    })

            elif pstate['state'] == STATE.IN_SEG:
                if c in (',', ')', '}', '\0'):
                    ident.add_segment(string[pstate['start']:pstate['i']].strip(),
                                       QUOTING.UNQUOTED,
                                       pstate['syntax'])
                    pstate['state'] = STATE.AFTER_SEG
                    return
                elif c == '.':
                    ident.add_segment(string[pstate['start']:pstate['i']].strip(),
                                       QUOTING.UNQUOTED,
                                       pstate['syntax'])
                    pstate.update({
                        'syntax': SYNTAX.NAME,
                        'state': STATE.BEFORE_SEG,
                        'i': pstate['i'] + 1
                    })
                elif c == '&':
                    ident.add_segment(string[pstate['start']:pstate['i']].strip(),
                                       QUOTING.UNQUOTED,
                                       pstate['syntax'])
                    pstate.update({
                        'syntax': SYNTAX.NEXT_KEY,
                        'state': STATE.BEFORE_SEG,
                        'i': pstate['i'] + 1
                    })

                else:
                    pstate['i'] += 1

            elif pstate['state'] == STATE.IN_BRACKET_SEG:
                if c == '\0':
                   raise ValueError("Expected ']', in member identifier %s" % string)
                if c == ']':
                    if string[pstate['i']+1] == ']':
                        pstate['i'] += 2
                        # fall through
                    else:
                        ident.add_segment(string[pstate['start']:pstate['i']+1].strip().replace(']]', ']'),
                                           QUOTING.QUOTED,
                                           pstate['syntax'])
                        pstate.update({
                            'i': pstate['i'] + 1,
                            'state': STATE.AFTER_SEG
                        })
                else:
                    pstate['i'] += 1
            elif pstate['state'] == STATE.AFTER_SEG:
                if c == ' ':
                    pstate['i'] += 1
                elif c == '.':
                    pstate.update({
                        'state': STATE.BEFORE_SEG,
                        'syntax': SYNTAX.NAME,
                        'i': pstate['i'] + 1
                    })
                elif c == '&':
                    pstate['state']= STATE.BEFORE_SEG
                    # Roll the syntax - NAME=>FIRST_KEY=>NEXT_KEY=>NEXT_KEY...
                    if pstate['syntax'] == SYNTAX.NAME:
                        pstate['syntax'] = SYNTAX.FIRST_KEY
                    else:
                        pstate['syntax'] = SYNTAX.NEXT_KEY
                    pstate['i'] += 1
                elif c == '\0':
                    pstate['i'] += 1
                    return
            else:
                raise AssertionError("Unexpected state: " + str(pstate))

        # this sucks, but lets me keep the algorithm
        # close to the original
        string += '\0'
        while pstate['i'] < k + 1:
            inner()

        if pstate['state'] == STATE.START:
            pass
        elif pstate['state'] == STATE.BEFORE_SEG:
            raise ValueError("Expected identifier after '.', in member identifier %s" % string)
        elif pstate['state'] == STATE.IN_BRACKET_SEG:
            ValueError("Expected ']' in member identifier %s" % string)

        return ident


if __name__ == '__main__':
    # TODO: port these tests
    # https://github.com/olap4j/olap4j/blob/b8eccd85753ffddb66c9d8b7c2cd7de2bd510ce0/testsrc/org/olap4j/impl/Olap4jUtilTest.java#L358
    print(Identifier.parse('[Date].[Year].[2010]').segments)
    print()
    print(Identifier.parse("[string].[with].[a [bracket]] in it]").segments)
    print()
    print(Identifier.parse("Time.1997.[Q3]").segments)
    print()
    print(Identifier.parse("[Customers].[City].&[San Francisco]&CA&USA.&[cust1234]").segments)
    print()
    print(Identifier.parse("[Date].Year").segments)
    print()



from operator import itemgetter

import sys
if sys.version_info >= (3, 0):
    from urllib.parse import urljoin
else: 
    from urlparse import urljoin

import copy
import requests

from .identifier import Identifier
from .aggregation import Aggregation

CUBE_ATTRS = ['name', 'dimensions', 'measures', 'annotations']
BOOL_OPTS = ['nonempty', 'distinct', 'parents']

class Cube(object):

    def __init__(self, name, dimensions, measures, annotations, client):
        self.name = name
        self.dimensions = dimensions
        self.measures = measures
        self.annotations = annotations
        self.client = client

        self._dimensions_by_name = {
            d['name']: d for d in self.dimensions
        }

    @property
    def time_dimension(self):
        tds = [d for d in self.dimensions if d['type'] == 'time']
        if len(tds) == 0:
            raise Exception("No time dimension defined in cube %s" % self.name)
        elif len(tds) > 1:
            raise Exception("More than 1 time dimension defined in cube %s" % self.name)
        return tds[0]

    @property
    def std_dimensions(self):
        """ Dict of non-time dimensions keyed by their name """
        return {
            d['name']: d
            for d in self.dimensions
            if d['type'] != 'time'
        }

    @property
    def dimensions_by_name(self):
        """ Dict of dimensions keyed by their name """
        return {
            d['name']: d
            for d in self.dimensions
        }

    @property
    def measures_by_name(self):
        return {
            m['name']: m
            for m in self.measures
        }

    def get_level(self, dimension_name, level_name, hierarchy=0):
        """
        Get level with name `level_name` in dimension with name `dimension_name`
        in hierarchy `hierarchy` (default: 0)
        """
        d = self._dimensions_by_name[dimension_name]
        for l in d['hierarchies'][hierarchy]['levels']:
            if l['name'] == level_name:
                return l

        # raise if not found
        raise ValueError('level with name `%s`` not found' % level_name)

    def get_members(self, dimension_name, level_name):
        return self.client.get_members(
            self.name,
            dimension_name,
            level_name
        )['members']

    def get_aggregation(self, drilldown=[], cut=[], measures=[], **extra_params):

        agg_params = copy.copy(extra_params)

        agg_params['drilldown'] = [
            self.get_level(*[seg.name for seg in Identifier.parse(dd).segments])
            for dd in drilldown
        ]

        agg_params['cut'] = cut

        agg_params['measures'] = [
            self.measures_by_name[m]
            for m in measures
        ]


        return self.client.get_aggregation(self,
                                           agg_params)


class MondrianClient(object):
    def __init__(self, api_base):
        self.api_base = api_base

    def get_cubes(self):
        r = self._request(urljoin(self.api_base, 'cubes')).json()
        return [Cube(*(itemgetter(*CUBE_ATTRS)(c) + (self,))) for c in r['cubes']]

    def get_cube(self, cube_id):
        r = self._request(urljoin(self.api_base, 'cubes/' + cube_id)).json()
        return Cube(*(itemgetter(*CUBE_ATTRS)(r) + (self,)))

    # TODO: validate shape of params
    def get_aggregation(self, cube, params):
        qs_params = {
            bo: 'true' if params.get(bo) else 'false'
            for bo in BOOL_OPTS
        }

        if len(params.get('measures', [])) == 0:
            raise Exception("Must provide at least one measure")

        for m in params['measures']:
            qs_params['measures[]'] = [m['name'] for m in params['measures']]

        if len(params.get('drilldown', [])) > 0:
            qs_params['drilldown[]'] = [l['full_name'] for l in params['drilldown']]

        if len(params.get('cut', [])) > 0:
            qs_params['cut[]'] = params['cut']

        if len(params.get('properties', [])) > 0:
            qs_params['properties[]'] = params['properties']

        if len(params.get('caption', [])) > 0:
            qs_params['caption[]'] = params['caption']

        r = self._request(
                urljoin(self.api_base, 'cubes/%s/aggregate' % cube.name),
                qs_params
            )

        return Aggregation(r.json(), cube, r.url, params)

    def get_members(self, cube_id, dimension, level):
        return self._request(
            urljoin(
                self.api_base,
                'cubes/%s/dimensions/%s/levels/%s/members' % (cube_id, dimension, level)
            )
        ).json()

    def get_member(self, cube, member_full_name):
        raise Exception('Not Implemented')

    def _request(self, url, params=None):
        return requests.get(url, params=params)



import unittest
from mock import patch, MagicMock, Mock
from operator import itemgetter
from urlparse import urljoin
import json
import os

import requests

from .client import MondrianClient, Cube, Aggregation, CUBE_ATTRS

API_BASE = 'http://mondrian'
FIXTURES_DIR =  os.path.join(os.path.dirname(os.path.realpath(__file__)), 'test_fixtures')

class TestMondrianClient(unittest.TestCase):

    def setUp(self):
        self.client = MondrianClient(API_BASE)
        with open(
                os.path.join(
                    FIXTURES_DIR,
                    'cube_response.json'
                )) as f:
            self.cube_fixture = f.read()

    @patch('requests.get')
    def test_get_cubes(self, MockRequests):
        MockRequests.json.return_value = { 'cubes': [json.loads(self.cube_fixture)] }
        cs = self.client.get_cubes()

        assert len(cs) == 1
        assert type(cs[0]) == Cube
        MockRequests.assert_called_with(urljoin(API_BASE, 'cubes'))

    @patch('requests.get')
    def test_get_one_cube(self, MockRequests):
        MockRequests.json.return_value = self.cube_fixture
        c = self.client.get_cube('foodmart')
        assert type(c) == Cube
        MockRequests.assert_called_with(urljoin(API_BASE, 'cubes/foodmart'))

    @patch('client.MondrianClient._request')
    def test_get_aggregation(self, mock_client_request):
        cube = json.loads(self.cube_fixture)
        c = self.client.get_aggregation('exports',
                                        {
                                            'drilldown': [cube['dimensions'][0]['hierarchies'][0]['levels'][1]],
                                            'measures': [cube['measures'][0]],
                                            'parents': 'true'
                                        })

        mock_client_request.assert_called_with(
            urljoin(API_BASE, 'cubes/exports/aggregate'),
            {
                'drilldown[]': ['[Date].[Year]'],
                'measures[]': ['Q Traded'],
                'nonempty': 'false',
                'parents': 'true',
                'distinct': 'false'
            }
        )

class TestAggregation(unittest.TestCase):
    def setUp(self):
        self.client = MondrianClient(API_BASE)
        with open(os.path.join(FIXTURES_DIR, 'aggregation_response.json')) as f:
            self.aggregation_fixture = json.load(f)

        with open(os.path.join(FIXTURES_DIR, 'aggregation_reponse_with_ancestors.json')) as f:
            self.aggregation_fixture_with_parents = json.load(f)

        with open(os.path.join(FIXTURES_DIR, 'cube_export.json')) as f:
            self.cube_response = json.load(f)

    def test_tidy_data(self):
        agg = Aggregation(self.aggregation_fixture)

        # TODO: add assertions
        print(len(agg.tidy['data']))


    def test_pandas_with_parents(self):
        cube = Cube(*(itemgetter(*CUBE_ATTRS)(self.cube_response) + (self.client,)))
        agg = Aggregation(self.aggregation_fixture_with_parents,
                          cube,
                          {'nonempty': True,
                           'drilldown': [{u'caption': u'Year', u'name': u'Year', u'full_name': u'[Date].[Year]'}, {u'caption': u'HS0', u'name': u'HS0', u'full_name': u'[HS].[HS2]'}],
                           'cut': [{'full_name': u'[Export Geography].[Region Metropolitana Santiago]'}],
                           'parents': True,
                           'measures': [{u'caption': u'FOB US', u'name': u'FOB US', u'full_name': u'[Measures].[FOB US]', u'annotations': {}}, {u'caption': u'Geo Rank Across Time', u'name': u'Geo Rank Across Time', u'full_name': u'[Measures].[Geo Rank Across Time]', u'annotations': {}}]})
        p = agg.to_pandas()

        #assert list(p.columns) == [u'Origin Country', u'Year', u'Exports']
        #assert len(p) == len(agg.tidy['data'])


if __name__ == '__main__':
    unittest.main()



from setuptools import setup

setup(
    name='mondrian-rest',
    version='0.2',
    description='Python client for mondrian-rest',
    url='http://github.com/Datawheel/mondrian-rest',
    author='Manuel Aristaran - Datawheel',
    author_email='manuel@jazzido.com',
    license='MIT',
    packages=['mondrian_rest'],
    install_requires=['numpy', 'pandas', 'requests', 'urlparse'],
    zip_safe=False)



from functools import reduce as reduce_
try:
    from itertools import product, izip, groupby
except ImportError:
    izip = zip

import numpy as np
import pandas as pd

from .identifier import Identifier


def parse_properties(properties):
    """
    parse an array of property specifications:

    input: ['ISICrev4.Level 2.Level 2 ES', 'ISICrev4.Level 1.Level 1 ES']
    output: {"ISICrev4"=>{"Level 2"=>["Level 2 ES"], "Level 1"=>["Level 1 ES"]}}
    """

    def reducer(h, it):
        k, v = it
        h[k] = dict(
            map(lambda jt: (jt[0], map(lambda vv: list(vv)[-1].name, jt[1])),
                groupby(v, key=lambda s: s.segments[1].name)))
        return h

    return reduce(
        reducer,
        groupby(
            sorted(
                map(Identifier.parse, properties),
                key=lambda p: ".".join(map(lambda s: s.name, p[:2]))),
            key=lambda s: s[0].name), {})


def get_props(row, pnames, props, dimensions):
    def reducer(h, row):
        ax_i, member = row
        dname = dimensions[ax_i]['name']
        if props.get(dname):
            mmbr_lvl = dimensions[ax_i]['level']
            for p in props[dname].get(mmbr_lvl, []):
                h[p] = member['properties'][p]
            if member.get('ancestors'):
                for l, p in filter(lambda it: it[0] != mmbr_lvl,
                                   props[dname].items()):
                    anc = next(anc for anc in member['ancestors']
                               if anc['level_name'] == l)
                    for prop in p:
                        h[prop] = anc['properties'][prop]

        return h

    pvalues = reduce(reducer, enumerate(row), {})

    return map(lambda pn: pvalues[pn], pnames)


class Aggregation(object):
    def __init__(self, data, cube, url, agg_params=None):
        self._data = data
        self._cube = cube
        self._agg_params = agg_params
        self.url = url

        self._tidy = None

    @property
    def axes(self):
        return self._data['axes']

    @property
    def measures(self):
        return self.axes[0]['members']

    @property
    def values(self):
        return self._data['values']

    @property
    def axis_dimensions(self):
        return self._data['axis_dimensions']

    @property
    def tidy(self):
        """ Unroll the JSON representation of a result into a
            'tidy' (http://vita.had.co.nz/papers/tidy-data.pdf) dataset. """

        if self._tidy is not None:
            return self._tidy

        data = self._data
        measures = data['axes'][0]['members']
        prod = [
            izip(e['members'], range(len(e['members'])))
            for e in data['axes'][1:]
        ]
        values = data['values']

        def build_row(cell):
            cidxs = list(reversed([c[1] for c in cell]))

            cm = [c[0] for c in cell]

            mvalues = [
                reduce_(
                    lambda memo, cur: memo[cur],  # navigate to values[coords]
                    cidxs + [mi],
                    values) for mi, m in enumerate(measures)
            ]

            return cm + mvalues

        self._tidy = {
            'axes': data['axis_dimensions'][1:],
            'measures': measures,
            'data': [build_row(cell) for cell in product(*prod)]
        }

        return self._tidy

    def to_pandas(self, filter_empty_measures=True):
        tidy = self.tidy
        columns = []
        table = []
        properties = self._agg_params.get('properties', [])
        measures = self._agg_params['measures']

        props = parse_properties(properties)
        pnames = [Identifier.parse(i).segments[-1].name for i in properties]

        # header row
        if self._agg_params['parents']:
            slices = []
            for dd in tidy['axes']:
                slices.append(dd['level_depth'])
                for ancestor_level in self._cube.dimensions_by_name[dd[
                        'name']]['hierarchies'][0]['levels'][1:dd[
                            'level_depth']]:
                    columns += [
                        'ID %s' % ancestor_level['caption'],
                        ancestor_level['caption']
                    ]
                columns += ['ID %s' % dd['level'], dd['level']]

            # property names
            columns += pnames

            # measure names
            columns += [m['caption'] for m in measures]

            for row in tidy['data']:
                r = []
                for j, cell in enumerate(row[:len(tidy['axes'])]):
                    for ancestor in reversed(
                            cell['ancestors'][:slices[j] - 1]):
                        r += [ancestor['key'], ancestor['caption']]
                    r += [cell['key'], cell['caption']]

                r += get_props(row[:-len(measures)], pnames, props,
                               tidy['axes'])

                for mvalue in row[len(tidy['axes']):]:
                    r.append(mvalue)

                table.append(r)

        else:  # no parents
            for dd in tidy['axes']:
                columns += ['ID %s' % dd['level'], dd['level']]
            # measure names
            columns += [m['caption'] for m in self._agg_params['measures']]

            for row in tidy['data']:
                r = []
                for cell in row[:len(tidy['axes'])]:
                    r += [cell['key'], cell['caption']]

                r += get_props(row[:-len(measures)], pnames, props,
                               tidy['axes'])

                for mvalue in row[len(tidy['axes']):]:
                    r.append(mvalue)

                table.append(r)

        df = pd.DataFrame(table,
                          columns=columns) \
               .set_index(columns[:-len(self._agg_params['measures'])])

        if filter_empty_measures:
            df = df[reduce_(
                np.logical_and,
                [df[msr['name']].notnull() for msr in self.measures])]

        return df
